package turtle.talk;

import java.io.Reader;
import java.io.StringReader;

/** Facade acts as a public interface to the turtle-talk package.
 * <p>Title: Turtle-talk</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: Memorial University</p>
 * @author Theodore Norvell
 * @version 1.0
 */
 
public class Facade {
    
    private Maze maze ;
    private Interpreter interpreter ;
    private static final int timeLimit = 50000 ;
    
    public Facade() {
        maze = new Maze() ;
        interpreter = new Interpreter(maze) ;
    }
    
    public Maze getMaze() { return maze ; }
    
    public void randomize() { maze.randomize(); } 
    
    public void compile( String program, StringSink stringSink ) {
    
        Reader reader = new StringReader( program ) ;
        TurtleTalkParser parser = new TurtleTalkParser( reader ) ;
        Analyser analyser = new Analyser() ;
        enterPredefinedItems( analyser ) ;
        parser.setAnalyser( analyser );
        interpreter.newProgram();
        parser.setCodeEmitter( interpreter );
        try {
            parser.Translation_Unit();
            analyser.done() ;
            stringSink.putString("Compilation successful\n");}
        catch( ParseException e ) {
            stringSink.putString( "Syntax Error\n"+e.getMessage() ); }
        catch( ConstraintError e ) {
            int line = parser.getToken(0).beginLine ;
            int column = parser.getToken(0).beginColumn ;
            stringSink.putString( "Constraint error near line "+line
                            +" column "+column+" :\n   "
                            +e.getMessage()+"\n" ); }
        catch( ResourceExhausted e ) {
            int line = parser.getToken(0).beginLine ;
            int column = parser.getToken(0).beginColumn ;
            stringSink.putString( "Sorry, resource contstraint exeeced near "+line
                            +" column "+column+" :\n   "
                            +e.getMessage() +"\n" ); }
        catch( AssertionError e ) {
            stringSink.putString(e.getMessage()+"\n" );
            e.printStackTrace(System.err); }
        catch( Throwable e ) {
            stringSink.putString("Internal Error: Unexpected exception\n");
            stringSink.putString(e.getMessage()+"\n" );
            e.printStackTrace(System.err); }
        return ; }
    
    
    public void run( final IntSource intSource,
                     final StringSink stringSink,
                     final IntSink intSink,
                     final boolean trace,
                     final int delay) {
        // To do: prevent more than one worker
        // thread from executing at once.
        // Also add support for stopping the
        // thread.
        Thread thread = new Thread() {
            public void run() {
                runPrivate( intSource, stringSink, intSink, trace, delay ) ; } } ;
        thread.start();
    } 
    
    private void runPrivate( IntSource intSource,
                     StringSink stringSink,
                     IntSink intSink,
                     boolean trace,
                     int delay) {
        try {
            interpreter.run( intSource, stringSink, intSink, trace, delay, timeLimit ) ;
            stringSink.putString("\nRun complete\n" ); }
        catch( TurtleTalkException e ) {
            stringSink.putString(e.getMessage()+"\n" ); }
        catch( AssertionError e ) {
            stringSink.putString(e.getMessage()+"\n" );
            e.printStackTrace(System.err); }
        catch( Throwable e ) {
            stringSink.putString("Internal Error: Unexpected exception\n");
            stringSink.putString(e.getMessage()+"\n" );
            e.printStackTrace(System.err); }
    }
    
    void enterPredefinedItems( Analyser analyser ) {
        final int DIR_TYPE = -1 ;
        /* direction */ {
            SymbolTableEntry directionEntry = new SymbolTableEntry(SymbolTableEntry.TYPE_ENTRY,
                                                             "direction", DIR_TYPE ) ;
            analyser.enterSymbol( directionEntry ) ; }
        /* up */ {
            CodeGenerationRule upCGR = new CodeGenerationRule() {
                public void apply( int numArgs, Analyser analyser, CodeEmitter codeEmitter )
                throws TurtleTalkException
                {
                    analyser.check(numArgs==0, "args after constant" );
                    codeEmitter.emitPush( Maze.UP );
                    analyser.push( DIR_TYPE ); } } ;
            
            SymbolTableEntry upEntry = new SymbolTableEntry( SymbolTableEntry.CONSTANT_ENTRY,
                                                               "up", upCGR ) ;
            analyser.enterSymbol( upEntry ) ; }
        /* right */ {
            CodeGenerationRule rightCGR = new CodeGenerationRule() {
                public void apply( int numArgs, Analyser analyser, CodeEmitter codeEmitter )
                throws TurtleTalkException
                {
                    analyser.check(numArgs==0, "args after constant" );
                    codeEmitter.emitPush( Maze.RIGHT );
                    analyser.push( DIR_TYPE ); } } ;
            
            SymbolTableEntry rightEntry = new SymbolTableEntry( SymbolTableEntry.CONSTANT_ENTRY,
                                                               "right", rightCGR ) ;
            analyser.enterSymbol( rightEntry ) ; }
        /* down */ {
            CodeGenerationRule downCGR = new CodeGenerationRule() {
                public void apply( int numArgs, Analyser analyser, CodeEmitter codeEmitter )
                throws TurtleTalkException
                {
                    analyser.check(numArgs==0, "args after constant" );
                    codeEmitter.emitPush( Maze.DOWN  );
                    analyser.push( DIR_TYPE ); } } ;
            
            SymbolTableEntry downEntry = new SymbolTableEntry( SymbolTableEntry.CONSTANT_ENTRY,
                                                               "down", downCGR ) ;
            analyser.enterSymbol( downEntry ) ; }
        /* left */ {
            CodeGenerationRule leftCGR = new CodeGenerationRule() {
                public void apply( int numArgs, Analyser analyser, CodeEmitter codeEmitter )
                throws TurtleTalkException
                {
                    analyser.check(numArgs==0, "args after constant" );
                    codeEmitter.emitPush( Maze.LEFT );
                    analyser.push( DIR_TYPE ); } } ;
            
            SymbolTableEntry leftEntry = new SymbolTableEntry( SymbolTableEntry.CONSTANT_ENTRY,
                                                               "left", leftCGR ) ;
            analyser.enterSymbol( leftEntry ) ; }
        /* procedure go( direction ) */ {
            CodeGenerationRule goCGR = new CodeGenerationRule() {
                public void apply( int numArgs, Analyser analyser, CodeEmitter codeEmitter )
                throws TurtleTalkException
                {
                    analyser.check(numArgs==1, "wrong number of arguments" );
                    analyser.check( analyser.top() == DIR_TYPE, "expected direction argument" ) ;
                    analyser.pop();
                    codeEmitter.emitByte( OpCode.GO ); } } ;
            
            SymbolTableEntry goEntry = new SymbolTableEntry( SymbolTableEntry.PROCEDURE_ENTRY,
                                                               "go", goCGR ) ;
            analyser.enterSymbol( goEntry ) ; }
        /* procedure putMark( direction ) */ {
            CodeGenerationRule putMarkCGR = new CodeGenerationRule() {
                public void apply( int numArgs, Analyser analyser, CodeEmitter codeEmitter )
                throws TurtleTalkException
                {
                    analyser.check(numArgs==1, "wrong number of arguments" );
                    analyser.check( analyser.top() == Analyser.INT_TYPE, "expected int argument" ) ;
                    analyser.pop();
                    codeEmitter.emitByte( OpCode.PUTMARK ); } } ;
            
            SymbolTableEntry putMarkEntry = new SymbolTableEntry( SymbolTableEntry.PROCEDURE_ENTRY,
                                                               "putMark", putMarkCGR ) ;
            analyser.enterSymbol( putMarkEntry ) ; }
        /* function wall( direction ) : bool */ {
            CodeGenerationRule wallCGR = new CodeGenerationRule() {
                public void apply( int numArgs, Analyser analyser, CodeEmitter codeEmitter )
                throws TurtleTalkException
                {
                    analyser.check(numArgs==1, "wrong number of arguments" );
                    analyser.check( analyser.top() == DIR_TYPE, "expected direction argument"  ) ;
                    analyser.pop();
                    codeEmitter.emitByte( OpCode.WALL );
                    analyser.push( Analyser.BOOL_TYPE );} } ;
            
            SymbolTableEntry wallEntry = new SymbolTableEntry( SymbolTableEntry.FUNCTION_ENTRY,
                                                               "wall", wallCGR ) ;
            analyser.enterSymbol( wallEntry ) ; }
        /* function goal() : bool */ {
            CodeGenerationRule goalCGR = new CodeGenerationRule() {
                public void apply( int numArgs, Analyser analyser, CodeEmitter codeEmitter )
                throws TurtleTalkException
                {
                    analyser.check(numArgs==0, "wrong number of arguments" );
                    codeEmitter.emitByte( OpCode.GOAL );
                    analyser.push( Analyser.BOOL_TYPE );} } ;
            
            SymbolTableEntry goalEntry = new SymbolTableEntry( SymbolTableEntry.FUNCTION_ENTRY,
                                                               "goal", goalCGR ) ;
            analyser.enterSymbol( goalEntry ) ; }
        /* function sniff() : direction */ {
            CodeGenerationRule sniffCGR = new CodeGenerationRule() {
                public void apply( int numArgs, Analyser analyser, CodeEmitter codeEmitter )
                throws TurtleTalkException
                {
                    analyser.check(numArgs==0, "wrong number of arguments" );
                    codeEmitter.emitByte( OpCode.SNIFF );
                    analyser.push( DIR_TYPE );} } ;
            
            SymbolTableEntry sniffEntry = new SymbolTableEntry( SymbolTableEntry.FUNCTION_ENTRY,
                                                               "sniff", sniffCGR ) ;
            analyser.enterSymbol( sniffEntry ) ; }
        /* function getMark() : int */ {
            CodeGenerationRule getMarkCGR = new CodeGenerationRule() {
                public void apply( int numArgs, Analyser analyser, CodeEmitter codeEmitter )
                throws TurtleTalkException
                {
                    analyser.check(numArgs==0, "wrong number of arguments" );
                    codeEmitter.emitByte( OpCode.GETMARK );
                    analyser.push( Analyser.INT_TYPE );} } ;
            
            SymbolTableEntry getMarkEntry = new SymbolTableEntry( SymbolTableEntry.FUNCTION_ENTRY,
                                                               "getMark", getMarkCGR ) ;
            analyser.enterSymbol( getMarkEntry ) ; }
        /* function next( direction ) : direction */ {
            CodeGenerationRule nextCGR = new CodeGenerationRule() {
                public void apply( int numArgs, Analyser analyser, CodeEmitter codeEmitter )
                throws TurtleTalkException
                {
                    analyser.check(numArgs==1, "wrong number of arguments" );
                    analyser.check( analyser.top() == DIR_TYPE, "expected direction argument" ) ;
                    analyser.pop();
                    codeEmitter.emitByte( OpCode.NEXT );
                    analyser.push( DIR_TYPE );} } ;
            
            SymbolTableEntry nextEntry = new SymbolTableEntry( SymbolTableEntry.FUNCTION_ENTRY,
                                                               "next", nextCGR ) ;
            analyser.enterSymbol( nextEntry ) ; }
    }
}