package turtle.talk;

/** An interpreter for the virtual machine code.
 *  <P> See OpCode and CodeEmitter documentation.
 * <p>Title: Turtle-talk</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: Mmorial University</p>
 * @author Theodore Norvell
 * @version 1.0
 */

class Interpreter implements CodeEmitter {
    
    // The state of the machine
    private static final int CODEMAX = 4096 ;
    private static final int STACKMAX = 128 ;
    /** The maximum number of variables allowed in scope at once.*/
    public static final int VARMAX = 128 ;
    private static final int STRINGMAX = 128 ;
    private boolean haveProgram = false ;
    private byte[] codeArray = new byte[CODEMAX] ;
    private int[] stack = new int[STACKMAX] ;
    private int[] vars = new int[VARMAX] ;
    private String[] string = new String[STRINGMAX] ;
    private int endOfCode = 0 ;
    private int nextStringAddress = 0 ;
    private int stackHeight = 0 ;
    private Maze maze ;
    
    private String[] opString = new String[128] ;
    
    public Interpreter(Maze maze) {
        this.maze = maze ;
        for( int i=0 ; i < 128 ; ++ i ) opString[i] = "Invalid" ;
        opString[ ADD ] = "ADD" ;
        opString[ SUBTRACT ] = "SUBTRACT" ;
        opString[ MULTIPLY ] = "MULTIPLY" ;
        opString[ DIVIDE ] = "DIVIDE" ;
        opString[ MOD ] = "MOD" ;
        opString[ EQUAL ] = "EQUAL" ;
        opString[ LESS ] = "LESS" ;
        opString[ LESSOREQUAL ] = "LESSOREQUAL" ;
        opString[ OR ] = "OR" ;
        opString[ NEGATE ] = "NEGATE" ;
        opString[ NOT ] = "NOT" ;
        opString[ PUSH16 ] = "PUSH" ;
        opString[ PUSH32 ] = "PUSH" ;
        opString[ FLIP ] = "FLIP" ;
        opString[ POP ] = "POP" ;
        opString[ FETCH ] = "FETCH" ;
        opString[ STORE ] = "STORE" ;
        opString[ BRANCH ] = "BRANCH" ;
        opString[ FBRANCH ] = "FBRANCH" ;
        opString[ PUTINT ] = "PUTINT" ;
        opString[ PUTSTRING ] = "PUTSTRING" ;
        opString[ GO ] = "GO" ;
        opString[ WALL ] = "WALL" ;
        opString[ GOAL ] = "GOAL" ;
        opString[ SNIFF  ] = "SNIFF" ;
        opString[ PUTMARK ] = "PUTMARK" ;
        opString[ GETMARK ] = "GETMARK" ;
        opString[ HALT ] = "HALT" ;
    }
    
    public void newProgram() { endOfCode = 0 ; haveProgram = true ; }
    
    public void emitByte( byte b ) throws ResourceExhausted {
        if( endOfCode == CODEMAX ) {
            throw new ResourceExhausted( "Limit Error: Program too long" ) ; }
        else {
            codeArray[ endOfCode ] = b ;
            endOfCode++ ; }
    }
    
    public void emitInt16( int i ) throws ResourceExhausted {
        emitByte( (byte) i ) ;
        emitByte( (byte)(i >>> 8 ) ) ;
    }
    
    public void emitInt32( int i ) throws ResourceExhausted {
        emitByte( (byte) i ) ;
        emitByte( (byte)(i >>> 8 ) ) ;
        emitByte( (byte)(i >>> 16 ) ) ;
        emitByte( (byte)(i >>> 24 ) ) ;
    }
    
    public void emitPush( int i )  throws ResourceExhausted {
        final int two_to_the_15 = (1<<15) ;
        if( 0 <= i && i < 128 ) {
            // Small positive ints are packed into the byte.
            emitByte( (byte)(-(i+1)) ) ; }
        else if( -two_to_the_15 <= i && i < two_to_the_15 ) {
            // Larger ints take 2 bytes + the instruction
            emitByte( PUSH16 ) ; emitInt16( i ) ; }
        else  {
            // Large ints take 4 bytes + the instruction
            emitByte( PUSH32 ) ; emitInt32( i ) ; }
    }
    
    public int currentInstructionAddress() { return endOfCode ; }
    
    public void patchByte( int address, byte b ) {
        Assert.assertTrue( 0 <= address && address < endOfCode, "bad address "+address ) ;
        codeArray[ address ] = b ; }
    
    public void patchInt16( int address, int i ) {
        Assert.assertTrue( 0 <= address && address < endOfCode-1, "bad address "+address ) ;
        codeArray[ address ] = (byte) i ;
        codeArray[ address+1 ] = (byte)(i >>> 8) ;
    }
    
    public int putString( String str ) throws ResourceExhausted {
        if( nextStringAddress < STRINGMAX ) {
            string[ nextStringAddress ] = str ;
            return nextStringAddress++ ; }
        else 
            throw new ResourceExhausted( "Limit Error: Too many strings" ) ;
    }
    
    private String getString( int address ) {
        Assert.assertTrue( 0 <= address && address < STRINGMAX, "bad string address "+address ) ;
        return string[ address ] ;
    }
    
    private int pop() {
        Assert.assertTrue( stackHeight > 0, "pop" ) ;
        return stack[ --stackHeight ] ;
    }
    
    private void push(int v) throws ResourceExhausted {
        if( stackHeight == STACKMAX ) throw new ResourceExhausted( "Limit Error: out of stack" ) ;
        stack[stackHeight++] = v ;
    }

    private void printStackTop( StringSink stringSink ) {
        int i = stackHeight ;
        while( i != 0 && i > stackHeight-3 ) {
            --i ;
            stringSink.putString( " "+stack[i] ) ; }
        if( i != 0 ) stringSink.putString( " ..." ) ;
        stringSink.putString("\n") ; }
        
    private int getVar( int address ) {
        Assert.assertTrue( 0 <= address && address < VARMAX, "Bad variable address "+address ) ;
        return vars[ address ] ;
    }
    
    private void putVar( int address, int value ) {
        Assert.assertTrue( 0 <= address && address < VARMAX, "Bad variable address "+address ) ;
        vars[ address ] = value ;
    }
    
    private byte getInstruction(int i ) {
        Assert.assertTrue( 0 <= i && i < CODEMAX, "code index out of bounds "+i ) ;
        return codeArray[ i] ;
    }
    
    private int getCodeInt16( int i ) {
            int lo= getInstruction( i ) ;
            int hi = getInstruction( i+1 ) ;
            return (hi<<8) | (0xFF&lo) ;
    }
    
    private int getCodeInt32( int i ) {
            int b0= getInstruction( i ) ;
            int b1 = getInstruction( i+1 ) ;
            int b2 = getInstruction( i+2 ) ;
            int b3 = getInstruction( i+3 ) ;
            return (b3<<24) | ((b2&0xFF)<<16) | ((b1&0xFF)<<8) | (b0&0xFF) ;
    }
    
    private void printInstruction(StringSink sink, int i) {
        byte op = getInstruction( i ) ;
        if( op < 0 ) {
            sink.putString( i+"\t: PUSHSMALL( "+ (-op - 1) +" )\n" ) ; }
        else if( op == PUSH16 ) {
            int j = getCodeInt16( i+1 ) ;
            sink.putString( i+"\t: PUSH16( "+ j +" )\n" ) ; }
        else if( op == PUSH32 ) {
            int j = getCodeInt32( i+1 ) ;
            sink.putString( i+"\t: PUSH32( "+ j +" )\n" ) ; }
        else {
            sink.putString( i+"\t: "+opString[ op ]+"\n") ; }
    }
    
    private boolean mazeTest(int direction, int value )  {
        int i, j ;
        switch( direction ) {
            case Maze.DOWN :
                i = 1 ; j = 0 ;
            break ; 
            case Maze.RIGHT :
                i = 0 ; j = 1 ;
            break ; 
            case Maze.UP :
                i = -1 ; j = 0 ;
            break ; 
            case Maze.LEFT :
                i = 0 ; j = -1 ;
            break ; 
            default :
                Assert.assertTrue( false, "bad direction" ) ;
                i=j=0 ;
        }
        int r = maze.getTurtleRow() ;
        int c = maze.getTurtleCol() ;
        return maze.get( r+i, c+j ) == value ;
    }
    
    private boolean turtleAtGoal() {
        int r = maze.getTurtleRow() ;
        int c = maze.getTurtleCol() ;
        return (maze.get( r, c ) & Maze.GOAL) != 0 ;
    }
    private void moveTurtle(int direction )  {
        int i, j ;
        switch( direction ) {
            case Maze.DOWN :
                i = 1 ; j = 0 ;
            break ; 
            case Maze.RIGHT :
                i = 0 ; j = 1 ;
            break ; 
            case Maze.UP :
                i = -1 ; j = 0 ;
            break ; 
            case Maze.LEFT :
                i = 0 ; j = -1 ;
            break ; 
            default :
                Assert.assertTrue( false, "bad direction" ) ;
                i=j=0 ;
        }
        int r = maze.getTurtleRow() ;
        int c = maze.getTurtleCol() ;
        if( maze.get(r+i,c+j) != Maze.WALL ) 
            maze.putTurtle( r+i, c+j ) ;
    }
    
    private int sniff() {
        int dr = maze.getGoalRow() - maze.getTurtleRow() ;
        int dc = maze.getGoalCol() - maze.getTurtleCol() ;
        if( Math.abs( dc ) > Math.abs( dr ) )
            return dc >= 0 ? Maze.RIGHT : Maze.LEFT ;
        else
            return dr >= 0 ? Maze.DOWN : Maze.UP ;
    }
                
    /** Run the program.
    * <p> If a program has been entered, it is executed until a
    * HALT instruction is encountered, an invalid instruction is encountered,
    * or time expires. There are also some resource limits that
    * can be exhausted: in particular the stack can overflow
    * @param intSource  An object to handle integer input
    * @param stringSink   An object for character output.
    * @param intSink    An object for integer output
    * @param trace      If true, a helpful trace will be sent to stringSink
    * @param limit      The maximum # of instructions to execute.
    * @see OpCode       Opcode
    */
    public void run(IntSource intSource,
                    StringSink stringSink, 
                    IntSink intSink,
                    boolean trace,
                    int delay,
                    long limit)
    throws ResourceExhausted {
        if( ! haveProgram ) {
            stringSink.putString("Can not run without first compiling") ;
            return ; }
        int pc = 0 ;
        stackHeight = 0 ;
        fetchExecuteLoop : while( true ) {
            limit -= 1 ;
            if( limit < 0 ) throw new ResourceExhausted( "Time limit exceeded. Possible infinite loop" ) ;
            if( trace ) {
                printStackTop( stringSink ) ; 
                printInstruction(stringSink, pc) ;}
            if( delay > 0 ) {
                long now = System.currentTimeMillis() ;
                while( System.currentTimeMillis() < now+delay ) 
                    try {
                        Thread.sleep( 1 ); }
                    catch( InterruptedException e ) { break ; }
            }
            byte op = getInstruction(pc) ;
            pc += 1 ;
            if( op < 0 ) {
                push( -op - 1 ) ; }
            else {
                switch( op ) {
                case ADD :
                case SUBTRACT :
                case MULTIPLY :
                case DIVIDE :
                case MOD :
                case EQUAL :
                case LESS :
                case LESSOREQUAL :
                case OR : {
                    int y = pop() ;
                    int x = pop() ;
                    int v ;
                    switch( op ) { 
                        case ADD : v = x+y ;
                        break ;
                        case SUBTRACT : v = x-y ;
                        break ;
                        case MULTIPLY : v = x*y ;
                        break ;
                        case DIVIDE : if( y==0 ) v=0 ; else v = x/y ;
                        break ;
                        case MOD :  if( y==0 ) v=0 ; else v = x%y ;
                        break ;
                        case EQUAL : if( x==y ) v = 1 ; else v = 0  ;
                        break ;
                        case LESS : if( x<y ) v = 1 ; else v = 0 ;
                        break ;
                        case LESSOREQUAL : if( x<=y ) v = 1 ; else v = 0 ;
                        break ;
                        case OR : if( x!=0 || y != 0 ) v=1 ; else v = 0 ;
                        break ;
                        default: // assert false ;
                            v = 0 ;
                    }
                    push( v ) ; }
                break ;
                case NEGATE : {
                    int x = pop() ;
                    push( -x ) ; }
                break ;
                case NOT : {
                    int x = pop() ;
                    push( x!=0 ? 0 : 1 ) ; }
                break ;
                case PUSH16 : {
                    int v = getCodeInt16( pc ) ;
                    pc += 2 ;
                    push( v ) ; }
                break ;
                case PUSH32 : {
                    int v = getCodeInt32( pc ) ;
                    pc += 4 ;
                    push( v ) ; }
                break ;
                case FLIP : {
                    int y = pop() ;
                    int x = pop() ;
                    push(y) ; push(x); }
                break ;
                case POP : {
                    pop() ; }
                break ;
                case FETCH : {
                    int x = pop() ;
                    push( getVar(x) ) ; }
                break ;
                case STORE : {
                    int y = pop() ;
                    int x = pop() ;
                    putVar( y, x ) ; }
                break ;
                case BRANCH : {
                    pc = pop() ; }
                break ;
                case FBRANCH : {
                    int y = pop() ;
                    int x = pop() ;
                    if( x == 0 ) pc = y ; }
                break ;
                case PUTINT : {
                    int x = pop() ;
                    intSink.putInt( x ) ; }
                break ;
                case PUTSTRING : {
                    int x = pop() ;
                    stringSink.putString( getString(x) ) ; }
                break ;
                case GETINT : {
                    int x = intSource.getInt() ;
                    push( x ) ; }
                break ;
                case GO : {
                    int x = pop() ;
                    moveTurtle( x ) ; }
                break ;
                case WALL : {
                    int x = pop() ;
                    if( mazeTest(x, Maze.WALL) ) push(1) ; else push(0) ; }
                break ;
                case GOAL : {
                    if( turtleAtGoal() ) push(1) ; else push(0) ; }
                break ;
                case SNIFF : {
                    int v = sniff() ;
                    push(v) ; }
                break ;
                case PUTMARK : {
                    int x = pop() ;
                    maze.putMark( x ) ; }
                break ;
                case GETMARK : {
                    int v = maze.getMark() ;
                    push( v) ; }
                break ;
                case NEXT : {
                    int x = pop() ;
                    push( x==3 ? 0 : x+1 ) ; }
                break ;
                case HALT :
                    break fetchExecuteLoop ;
                default:
                    Assert.assertTrue( false, "bad instruction" ) ;
            } } }
        if( trace ) { stringSink.putString( "Program terminated\n" ) ; }
    }   
}