package turtle.talk;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import javax.swing.event.SwingPropertyChangeSupport;

/** Model for the maze.
 * <p>Title: Turtle-talk</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: Mmorial University</p>
 * @author Theodore Norvell
 * @version 1.0
 */

public class Maze {

    public static final int BLANK=0, WALL=1, TURTLE=2, GOAL=4 ;
    
    public static final int 
        // The four directions are encoded as
        UP = 0,
        RIGHT = 1,
        DOWN = 2 ,
        LEFT = 3 ;
    private int width, height ;
    private int[][] a ;
    // Invariant: a is rectangular with height rows and width columns.
    // Invariant: a[i][j] is one of of {BLANK, WALL, TURTLE, GOAL, TURTLE|GOAL},
    //                  for all i in {0,..height}, j in {0,..width}
    private int[][] mark ;
    // Invariant: mark is the same shape as a
    private int turtleRow, turtleCol, goalRow, goalCol ;
    // Invariant: a[turtleRow][turtleCol] contains TURTLE, and only that square contains TURTLE
    // Invariant: a[goalRow][goalCol] contains GOAL, and only that square contains GOAL
    
    private java.util.Random  rng = new java.util.Random() ;
    
    private double[] probTable = new double[]{
        // 0     1     2     3     4     5     6      7     8
        0.00, 0.70, 0.20, 0.10, 0.05, 0.05, 0.90,  0.90,  1.0 } ;
    
    private SwingPropertyChangeSupport changeSupport ;
    
    public Maze() {
        changeSupport = new SwingPropertyChangeSupport(this) ;
        randomize() ;
    }
    
    public synchronized int getHeight() { return height ; }
    
    public synchronized int getWidth() { return width ; }
    
    public synchronized int getTurtleRow() { return turtleRow ; }
    
    public synchronized int getTurtleCol() { return turtleCol ; }
    
    public synchronized int getGoalRow() { return goalRow ; }
    
    public synchronized int getGoalCol() { return goalCol ; }
    
    public synchronized int getMark() {
        return mark[turtleRow][turtleCol] ; }
    
    public synchronized int get(int i, int j) {
        // assert( 0 <=i && i < height && 0 <= j && j < width ) ;
        return a[i][j] ; }
    
    public synchronized void putTurtle( int i, int j ) {
        // assert( 0 <=i && i < height && 0 <= j && j < width ) ;
        // assert( a[i][j] == BLANK ) ;
        a[turtleRow][turtleCol] = a[turtleRow][turtleCol] & ~TURTLE ;
        turtleRow = i ;
        turtleCol = j ;
        a[turtleRow][turtleCol] = a[turtleRow][turtleCol] | TURTLE ;
        fire() ; }
    
    public synchronized void putMark( int v ) {
        mark[turtleRow][turtleCol] = v ; }
        
    public synchronized void randomize() {
        height = 20 ;
        width = 20 ;
        a = new int[height][width] ;
        
        // Blank every square
            for( int i=0 ; i<height ; ++i )
                for( int j=0 ; j<width ; ++j ) a[i][j] = BLANK ;
        // Walls around the perimeter
            for( int i=0 ; i<height ; ++i ) a[i][0] = a[i][width-1] = WALL ;
            for( int j=0 ; j<width ; ++j ) a[0][j] = a[height-1][j] = WALL ;
        
        mark = new int[height][width] ;
        
        // Place the turtle on a blank
            turtleRow = rand(1, height-2) ;
            turtleCol = rand(1, width-2) ;
            a[turtleRow][turtleCol] =  TURTLE ;
        
        // Fill the remaining blanks with a random wall pattern
        // (Note really should ensure that there is a path.)
            for( int k = 0 ; k < 50 ; ++k )
                for( int i=1 ; i<height-1 ; ++i )
                    for( int j=1 ; j<width-1 ; ++j ) 
                        if( a[i][j] == BLANK || a[i][j] == WALL) {
                            // Count the walls in the neighbourhood of 8
                            int count = a[i][j] == WALL ? -1 : 0 ;
                            for( int x=-1 ; x <= +1 ; ++x ) 
                                for( int y=-1 ; y <= +1 ; ++y )
                                    if( a[i+x][j+y] == WALL ) count++ ;
                            if( Math.random() < probTable[count] )
                                a[i][j] = WALL ;
                            else 
                                a[i][j] = BLANK ; }
        
        // Place the goal by performing a random walk from the turtle position.
            goalRow = turtleRow ;
            goalCol = turtleCol ;
            for( int k = 0 ; k < (height+width)*10 ; ++k ) {
                int r = rand( 0, 4 ) ;
                int dr = r==0 ? -1 : r==2 ? +1 : 0 ;
                int dc = r==1 ? -1 : r==3 ? +1 : 0 ;
                if( a[goalRow + dr][goalCol + dc] == BLANK ) {
                    goalRow += dr ;
                    goalCol += dc ; } }
            a[goalRow][goalCol] = a[goalRow][goalCol] | GOAL ;
        
        fire() ;
    }
    
    private int rand( int offset, int size ) {
        int k ;
        do { k = Math.abs( rng.nextInt() ) ; 
        } while( k<0 ) ;
        return offset + k % size ;}
    
    public synchronized void addListener( PropertyChangeListener listener ) {
        changeSupport.addPropertyChangeListener( listener ); }
    
    private synchronized void fire() {
        changeSupport.firePropertyChange( new PropertyChangeEvent(this, "", null, null) ) ; }        
}