package turtle.talk;

import junit.framework.*;
import java.io.*;

abstract public class AbstractTurtleTalkTest extends TestCase {
    
    protected static final int SUCCESS = 0 ;
    protected static final int SYNTAX = 1 ;
    protected static final int CONSTRAINT = 2 ;
    protected static final int RESOURCE = 3 ;
    protected static final int OTHER = 4 ;
    
    protected boolean print = true;

    protected Maze maze = new Maze() ;
    protected Interpreter interpreter = new Interpreter(maze) ;
    protected TestInput noTestInput
        = new TestInput( new int[] {} ) ;
    protected ExpectedOutput noExpectedOutput
        = new ExpectedOutput( new int[] {} ) ;
    protected ExpectedStringOutput noExpectedStringOutput
        = new ExpectedStringOutput( new String[] {} ) ;
    protected DontCareOutput dontCare
        = new DontCareOutput( ) ;
    
    public AbstractTurtleTalkTest(String name) { super (name); }

    protected void put( String message ) {
        if( print ) System.out.println( message ) ; }
        
    protected void setUp () {
        put ( "--------------------- Start Test in "+this.getName()+" --------------------------" ) ; }
    
    public void tearDown() {
        put( "******************End Test******************" ) ; }

    protected void compileFileExpectingSuccess( String fileName ) {
        int success = compileFile( fileName ) ;
        if( success != SUCCESS ) put( "Expected success" ) ;
        assertTrue( "Compilation failed", success == SUCCESS ) ;
    }

    protected void compileFileExpectingConstraint( String fileName ) {
        int success = compileFile( fileName ) ;
        if( success != CONSTRAINT ) put( "Expected constraint failure" ) ;
        assertTrue( "Constraint error expected", success == CONSTRAINT ) ;
    }

    protected void compileFileExpectingSyntax( String fileName ) {
        int success = compileFile( fileName ) ;
        if( success != SYNTAX ) put( "Expected syntax error" ) ;
        assertTrue( "Expected syntax error", success == SYNTAX ) ;
    }

    protected int compileFile( String fileName ) {
        int success;
        Reader reader ;
        try { reader = new FileReader( fileName ) ; }
        catch( FileNotFoundException e ) {
            success = OTHER ;
            put( "File "+fileName+" not found" ) ;
            return success ; 
        }
        TurtleTalkParser parser = new TurtleTalkParser( reader ) ;
        Analyser analyser = new Analyser() ;
        //enterPredefinedItems( analyser ) ;
        parser.setAnalyser( analyser );
        interpreter.newProgram();
        parser.setCodeEmitter( interpreter );
        try {
            parser.Translation_Unit();
            analyser.done() ;
            put("Compilation successful\n");
            success = SUCCESS ; }
        catch( ParseException e ) {
            success = SYNTAX ;
            put( "Syntax Error\n"+e.getMessage() ); }
        catch( ConstraintError e ) {
            success = CONSTRAINT ;
            int line = parser.getToken(0).beginLine ;
            int column = parser.getToken(0).beginColumn ;
            put( "Constraint error near line "+line
                            +" column "+column+" :\n   "
                            +e.getMessage()+"\n" ); }
        catch( ResourceExhausted e ) {
            success = RESOURCE ;
            int line = parser.getToken(0).beginLine ;
            int column = parser.getToken(0).beginColumn ;
            put( "Sorry, resource contstraint exeeced near "+line
                            +" column "+column+" :\n   "
                            +e.getMessage() +"\n" ); }
        catch( AssertionError e ) {
            success = OTHER ;
            put(e.getMessage()+"\n" );
            if( print ) e.printStackTrace(System.out); }
        catch( Throwable e ) {
            success = OTHER ;
            put( "Internal Error: Unexpected exception\n" );
            put( e.getMessage()+"\n" );
            if( print ) e.printStackTrace(System.out); }
        return success ; }
        
        
    protected void run( IntSource intSource,
                     StringSink stringSink,
                     IntSink intSink) {
        int success ;
        boolean trace = false ;
        int timeLimit = 10000 ;
        try {
            interpreter.run( intSource, stringSink, intSink, trace, 0, timeLimit ) ;
            intSource.close();
            stringSink.close();
            intSink.close();
            success = SUCCESS ; }
        catch( junit.framework.AssertionFailedError e ) {
            throw e ; }
        catch( TurtleTalkException e ) {
            success = RESOURCE ;
            put(e.getMessage()+"\n" ); }
        catch( AssertionError e ) {
            success = OTHER ;
            put(e.getMessage()+"\n" );
            if( print ) e.printStackTrace(System.out); }
        catch( Throwable e ) {
            success = OTHER ;
            put("Internal Error: Unexpected exception\n");
            put(e.getMessage()+"\n" );
            if( print ) e.printStackTrace(System.out);  }
        assertTrue( "Run failed", success==SUCCESS ) ;
    }
}